from __future__ import annotations

import libcst
import libcst.matchers

from ._base import ErrorCollectingVisitor

MULTILINE_SPACE_MATCHER = (
    libcst.matchers.SimpleWhitespace(
        value=libcst.matchers.MatchIfTrue(lambda x: "\n" in x)
    )
    | libcst.matchers.ParenthesizedWhitespace()
)


class AnnotationWrapErrorCollector(ErrorCollectingVisitor):
    METADATA_DEPENDENCIES = (libcst.metadata.PositionProvider,)

    def visit_Param(self, node: libcst.Param) -> None:
        if libcst.matchers.matches(
            node,
            libcst.matchers.Param(
                annotation=libcst.matchers.Annotation(
                    annotation=libcst.matchers.BinaryOperation(
                        lpar=[],
                        rpar=[],
                    )
                )
            ),
        ):
            leftmost_binop: libcst.BinaryOperation = (
                node.annotation.annotation  # type: ignore[assignment, union-attr]
            )
            while isinstance(leftmost_binop.left, libcst.BinaryOperation):
                leftmost_binop = leftmost_binop.left
            if libcst.matchers.matches(
                leftmost_binop.operator,
                libcst.matchers.BitOr(whitespace_before=MULTILINE_SPACE_MATCHER),
            ) or libcst.matchers.matches(
                leftmost_binop.operator,
                libcst.matchers.BitOr(whitespace_after=MULTILINE_SPACE_MATCHER),
            ):
                lpos = self.get_metadata(
                    libcst.metadata.PositionProvider, leftmost_binop
                ).start
                self.errors.add((lpos.line, self.filename, "W120"))



import os

from slyp.fixers import fix_file


def test_fixer_no_op(capsys, tmpdir):
    os.chdir(tmpdir)
    tmpdir.join("foo.py").write(
        """\
x = 1
"""
    )
    res = fix_file("foo.py", verbose=True)

    assert res
    assert "slyp: no changes to foo.py" in capsys.readouterr().out
    assert tmpdir.join("foo.py").read() == "x = 1\n"


def test_fixer_is_no_op_on_needed_parens(tmpdir):
    os.chdir(tmpdir)
    body = """\
a = [(1,), (2, 3)]
b = foo((x % 2 for x in range(10)), bar)
c = (1+2)*3
d = -1 * (-(1+2))
e = (foo() if x else bar())["baz"]
f = (await foo())["bar"]
g = (1).bit_length()
h = (2.2).is_integer()
i = (x is y) in truefalsecontainer
"""
    tmpdir.join("foo.py").write(body)
    res = fix_file("foo.py", verbose=True)
    assert res
    assert tmpdir.join("foo.py").read() == body


def test_fix_unnecessary_double_paren_tuple(capsys, tmpdir):
    os.chdir(tmpdir)
    tmpdir.join("foo.py").write(
        """\
a = ((1, 2))
"""
    )
    res = fix_file("foo.py", verbose=True)

    assert not res
    assert "slyp: fixing foo.py" in capsys.readouterr().out
    assert tmpdir.join("foo.py").read() == "a = (1, 2)\n"


def test_fix_unnecessary_many_paren_string(capsys, tmpdir):
    os.chdir(tmpdir)
    tmpdir.join("foo.py").write(
        """\
a = ((((((("foo")))))))
"""
    )
    res = fix_file("foo.py", verbose=True)

    assert not res
    assert "slyp: fixing foo.py" in capsys.readouterr().out
    assert tmpdir.join("foo.py").read() == 'a = "foo"\n'


def test_paren_fixer_preserves_innermost_under_splatarg(capsys, tmpdir):
    os.chdir(tmpdir)
    tmpdir.join("foo.py").write(
        """\
foo(*(("a b c".split())))
"""
    )
    res = fix_file("foo.py", verbose=False)

    assert not res
    assert "slyp: fixing foo.py" in capsys.readouterr().out
    assert (
        tmpdir.join("foo.py").read()
        == """\
foo(*("a b c".split()))
"""
    )



from __future__ import annotations

import re

from slyp.codes import CODE_MAP

from .abstract import run_ast_checkers
from .concrete import run_cst_checkers

_DISALBE_RE = re.compile(rb"#\s*slyp:\s*disable=(.*)")


def check_file(
    filename: str, *, verbose: bool, disabled_codes: set[str], enabled_codes: set[str]
) -> bool:
    if verbose:
        print(f"checking {filename}")

    with open(filename, "rb") as fp:
        bin_data = fp.read()

    try:
        cst_errors = run_cst_checkers(filename, bin_data)
    except RecursionError:
        cst_errors = {(0, "X002")}

    ast_errors = run_ast_checkers(filename, bin_data)
    errors = sorted(cst_errors | ast_errors)

    lines = bin_data.splitlines()
    filtered_errors = sorted(
        (lineno, code)
        for lineno, code in errors
        if not _disabled(code, disabled_codes, enabled_codes)
        and not _exempt(lines, lineno - 1, code)
    )

    if filtered_errors:
        for lineno, code in filtered_errors:
            print(f"{filename}:{lineno}: {CODE_MAP[code]}")
        return False
    return True


def _disabled(code: str, disabled_codes: set[str], enabled_codes: set[str]) -> bool:
    cdef = CODE_MAP[code]

    # enabled is higher precedence than disabled
    # but "all" in enabled_codes is not checked so that
    # '--enable all --disable FOO' works
    if code in enabled_codes or cdef.category in enabled_codes:
        return False

    # if "all" is in disabled_codes, then everything is disabled unless
    # explicitly enabled (checked above)
    if (
        "all" in disabled_codes
        or code in disabled_codes
        or cdef.category in disabled_codes
    ):
        return True

    return False


def _exempt(lines: list[bytes], lineno: int, code: str) -> bool:
    # should be impossible most of the time, but a failed parse uses a position of
    # 0 which becomes -1 here
    if lineno == -1 or len(lines) < lineno:
        return False

    line = lines[lineno]
    # start with a faster check before trying the regex
    if b"slyp" not in line:
        return False

    if match := _DISALBE_RE.search(line):
        disabled_codes = match.group(1)
        return disabled_codes == b"all" or code.encode() in disabled_codes.split(b",")

    return False


def _clear_errors() -> None:
    # testsuite hook for resetting errors for clean reporting
    from .abstract import _clear_visitor_errors as clear_cst_errors
    from .concrete import _clear_visitor_errors as clear_ast_errors

    clear_cst_errors()
    clear_ast_errors()



from __future__ import annotations

import argparse
import glob
import os
import stat
import subprocess
import sys
import textwrap
import typing as t

from slyp.checkers import check_file
from slyp.codes import CODE_MAP
from slyp.fixers import fix_file

DEFAULT_DISABLED_CODES: set[str] = {"W201", "W202", "W203"}


def main() -> None:
    parser = argparse.ArgumentParser(
        description="slyp is a linter and fixer for Python code",
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )
    parser.add_argument(
        "--list", action="store_true", help="list all error and warning codes"
    )
    parser.add_argument(
        "-v", "--verbose", action="store_true", help="increase output verbosity"
    )
    parser.add_argument(
        "--use-git-ls", action="store_true", help="find python files from git-ls-files"
    )
    parser.add_argument(
        "--disable",
        help="Disable error and warning codes (comma delimited)",
        default="",
    )
    parser.add_argument(
        "--enable",
        help=(
            "Enable error and warning codes which are otherwise disabled "
            "(comma delimited, overrides --disable)"
        ),
        default="",
    )
    parser.add_argument("files", nargs="*", help="default: all python files")
    args = parser.parse_args()

    if args.list:
        list_codes()
        sys.exit(0)

    if args.use_git_ls and args.files:
        parser.error("--use-git-ls requires no filenames as arguments")

    # parse inputs from comma delimited lists
    disabled_codes = {x for x in args.disable.split(",") if x != ""}
    enabled_codes = {x for x in args.enable.split(",") if x != ""}
    # add default disables if "all" is not in --enable
    if "all" not in enabled_codes:
        disabled_codes = disabled_codes | DEFAULT_DISABLED_CODES

    success = True
    for filename in all_py_filenames(args.files, args.use_git_ls):
        success = (
            fix_file(filename, verbose=args.verbose)
            and check_file(
                filename,
                verbose=args.verbose,
                disabled_codes=disabled_codes,
                enabled_codes=enabled_codes,
            )
            and success
        )

    if not success:
        sys.exit(1)

    if args.verbose:
        print("ok")


def all_py_filenames(files: t.Sequence[str], use_git_ls: bool) -> t.Iterable[str]:
    if files:
        yield from files
    elif use_git_ls:
        git_ls_files_proc = subprocess.run(
            ["git", "ls-files"], check=True, capture_output=True, text=True
        )
        candidates = git_ls_files_proc.stdout.split("\n")
        for file in candidates:
            if file == "":
                continue
            if is_python(file):
                yield file
    else:
        yield from glob.glob("**/*.py", recursive=True)


def is_python(filename: str) -> bool:
    # don't pay attention to symlinks or other special filetypes
    # those aren't really files to check
    # but you might be fed them by `git ls-files`
    if not stat.S_ISREG(os.lstat(filename).st_mode):
        return False

    # .py is good
    if filename.endswith(".py"):
        return True

    # if it's not executable, it couldn't be a script
    if not os.access(filename, os.X_OK):
        return False

    # but otherwise, look for a shebang which contains 'python' as a substring
    try:
        with open(filename, encoding="utf-8") as fp:
            firstline = fp.readline()
    except UnicodeDecodeError:
        return False
    if firstline.startswith("#!") and "python" in firstline:
        return True

    return False


def list_codes() -> None:
    first = True
    for code in CODE_MAP.values():
        if code.hidden:
            continue
        if first:
            first = False
        else:
            print()

        if code.default_disabled:
            description: str = f"{code.message} (disabled by default)"
        else:
            description = code.message
        print(f"{code.code}: {description}")
        print(textwrap.indent(code.example, "    "))



from __future__ import annotations

import ast

from ._base import ErrorRecordingVisitor


class NoneCheckedVarReturnVisitor(ErrorRecordingVisitor):
    def visit_If(self, node: ast.If) -> None:
        if _is_var_none_check(node.test):
            var_name = node.test.left.id  # type: ignore[attr-defined]
            if _is_return_var(node.body, var_name):
                self.errors.add((node.lineno, self.filename, "E110"))
                return
        self.generic_visit(node)


def _is_var_none_check(node: ast.expr) -> bool:
    return (
        isinstance(node, ast.Compare)
        and isinstance(node.left, ast.Name)
        and len(node.ops) == 1
        and isinstance(node.ops[0], ast.Is)
        and len(node.comparators) == 1
        and isinstance(node.comparators[0], ast.Constant)
        and node.comparators[0].value is None
    )


def _is_return_var(body: list[ast.stmt], var_name: str) -> bool:
    return (
        len(body) == 1
        and isinstance(body[0], ast.Return)
        and isinstance(body[0].value, ast.Name)
        and body[0].value.id == var_name
    )



#!/usr/bin/env python
import re
import sys


def get_old_version():
    with open("pyproject.toml") as fp:
        content = fp.read()
    match = re.search(r'^version = "(\d+\.\d+\.\d+)"$', content, flags=re.MULTILINE)
    if not match:
        raise ValueError("did not find old version in pyproject.toml")
    return match.group(1)


def replace_version(filename, prefix, old_version, new_version):
    print(f"updating {filename}")
    with open(filename) as fp:
        content = fp.read()
    old_str = prefix + old_version
    new_str = prefix + new_version
    content = content.replace(old_str, new_str)
    with open(filename, "w") as fp:
        fp.write(content)


def update_changelog(new_version):
    print("updating CHANGELOG.md")
    with open("CHANGELOG.md") as fp:
        content = fp.read()

    marker = "<!-- changelog-unreleased-marker -->"

    content = content.replace(
        marker,
        f"""{marker}

# {new_version}""",
    )
    with open("CHANGELOG.md", "w") as fp:
        fp.write(content)


def parse_version(s):
    vals = s.split(".")
    if len(vals) != 3:
        raise ValueError(f"invalid version string: {s}")
    return tuple(int(x) for x in vals)


def comparse_versions(old_version, new_version):
    if parse_version(new_version) <= parse_version(old_version):
        raise ValueError("new version must be greater than old version")


def main():
    if len(sys.argv) != 2:
        sys.exit(2)

    new_version = sys.argv[1]
    old_version = get_old_version()
    print(f"old = {old_version}, new = {new_version}")
    comparse_versions(old_version, new_version)

    replace_version("pyproject.toml", 'version = "', old_version, new_version)
    replace_version("README.md", "rev: ", old_version, new_version)
    update_changelog(new_version)


if __name__ == "__main__":
    main()



from __future__ import annotations

import typing as t

import libcst
import libcst.matchers

# alphabetical order
NODE_TYPES = t.Union[
    libcst.Attribute,
    libcst.Await,
    libcst.BinaryOperation,
    libcst.Call,
    libcst.Comparison,
    libcst.ConcatenatedString,
    libcst.Dict,
    libcst.DictComp,
    libcst.Ellipsis,
    libcst.Float,
    libcst.FormattedString,
    libcst.GeneratorExp,
    libcst.IfExp,
    libcst.Imaginary,
    libcst.Integer,
    libcst.Lambda,
    libcst.List,
    libcst.ListComp,
    libcst.Name,
    libcst.Set,
    libcst.SetComp,
    libcst.SimpleString,
    libcst.StarredElement,
    libcst.Subscript,
    libcst.Tuple,
    libcst.UnaryOperation,
    libcst.Yield,
]
MATCH_ON = (
    getattr(libcst.matchers, type_.__name__)() for type_ in t.get_args(NODE_TYPES)
)

# The innermost parens should not be removed in some cases, as they are
# required in some contexts.
PRESERVE_INNERMOST_PAREN_TYPES: tuple[type[libcst.CSTNode], ...] = (
    # list of tuples:
    #   [(1, 2), (3, 4)] != [1, 2, 3, 4]
    libcst.Tuple,
    # generator expression passed to a function call:
    #   foo((x % 2 for x in range(10)), bar)
    libcst.GeneratorExp,
    # parens define precedence:
    #   (1 + 2) * 3 != 1 + 2 * 3
    libcst.BinaryOperation,
    libcst.UnaryOperation,
    # precedence also for comparisons:
    #   (x is y) in truefalsecontainer
    libcst.Comparison,
    # parens are required for usage of the results:
    #   (foo() if x else bar())["baz"]
    #   (await foo())["bar"]
    libcst.IfExp,
    libcst.Await,
    # direct access to attributes of integers and floats requires parens:
    #   (1).bit_length()
    libcst.Integer,
    libcst.Float,
)


class UnnecessaryParenthesesFixer(libcst.matchers.MatcherDecoratableTransformer):
    METADATA_DEPENDENCIES = (
        libcst.metadata.PositionProvider,
        libcst.metadata.ParentNodeProvider,
    )

    def _how_many_parens_same_line(self, node: libcst.CSTNode) -> int:
        lpar, rpar = node.lpar, node.rpar  # type: ignore[attr-defined]

        max_offset = -1
        for offset in range(len(lpar)):
            lpar_line = self.get_metadata(
                libcst.metadata.PositionProvider, lpar[-(offset + 1)]
            ).start.line
            rpar_line = self.get_metadata(
                libcst.metadata.PositionProvider, rpar[offset]
            ).start.line

            if lpar_line != rpar_line:
                break
            max_offset = offset
        return max_offset + 1

    def _parent_node_is_argsplat(self, node: libcst.CSTNode) -> bool:
        """check if the parent of the node is *-expansion of an arg"""
        parent = self.get_metadata(libcst.metadata.ParentNodeProvider, node)
        if not isinstance(parent, libcst.Arg):
            return False
        return bool(parent.star)

    @libcst.matchers.leave(libcst.matchers.OneOf(*MATCH_ON))
    def modify_parenthesized_node(
        self, original_node: NODE_TYPES, updated_node: NODE_TYPES
    ) -> NODE_TYPES:
        if not original_node.lpar:
            return updated_node

        preserve_innermost = isinstance(original_node, PRESERVE_INNERMOST_PAREN_TYPES)
        if not preserve_innermost and self._parent_node_is_argsplat(original_node):
            preserve_innermost = True

        if preserve_innermost and len(original_node.lpar) == 1:
            return updated_node

        num_parens_to_unwrap = self._how_many_parens_same_line(original_node)
        if preserve_innermost:
            num_parens_to_unwrap -= 1
        if num_parens_to_unwrap <= 0:
            return updated_node
        return updated_node.with_changes(
            lpar=updated_node.lpar[:-num_parens_to_unwrap],
            rpar=updated_node.rpar[num_parens_to_unwrap:],
        )



from __future__ import annotations

import libcst

from .unnecessary_parens import UnnecessaryParenthesesFixer

_FIXERS: list[libcst.CSTTransformer] = [
    UnnecessaryParenthesesFixer(),
]


def fix_file(filename: str, *, verbose: bool) -> bool:
    """returns True if no changes were needed"""
    with open(filename, "rb") as fp:
        bin_data = fp.read()

    try:
        new_data = _fix_data(bin_data)
    # ignore failures to parse and treat these as "unchanged"
    # linting will flag these independently
    except (RecursionError, libcst.ParserSyntaxError):
        new_data = bin_data

    if new_data == bin_data:
        if verbose:
            print(f"slyp: no changes to {filename}")

        return True

    print(f"slyp: fixing {filename}")
    with open(filename, "wb") as fp:
        fp.write(new_data)

    return False


def _fix_data(content: bytes) -> bytes:
    raw_tree = libcst.parse_module(content)
    tree = libcst.MetadataWrapper(raw_tree)

    for fixer in _FIXERS:
        tree = tree.visit(fixer)  # type: ignore[assignment]

    code: str = tree.code  # type: ignore[attr-defined]
    return code.encode("utf-8")



import os

import pytest

from slyp.checkers import _clear_errors, check_file


@pytest.fixture(autouse=True)
def _auto_clear_errors():
    _clear_errors()


def test_check_passes_on_example_file(tmpdir):
    os.chdir(tmpdir)
    tmpdir.join("foo.py").write(
        """\
x = "foo bar"

foo(x="foo bar")
foo(
    x=(
        "foo "
        "bar"
    )
)

y = {
    "foo": "bar",
}
y = {
    "foo": ("bar"
    "baz"),
}

z = (
    "alpha",
    "beta",
    "gamma",
)
z = [
    "alpha",
    "beta",
    "gamma",
]
z = {
    "alpha",
    "beta",
    "gamma",
}
"""
    )
    res = check_file("foo.py", verbose=False, disabled_codes=set(), enabled_codes=set())
    assert res is True


def test_check_captures_e100(tmpdir, capsys):
    os.chdir(tmpdir)
    tmpdir.join("foo.py").write(
        """\
x = "foo" "bar"
"""
    )
    res = check_file("foo.py", verbose=False, disabled_codes=set(), enabled_codes=set())
    assert res is False

    assert "foo.py:1: unnecessary string concat (E100)" in capsys.readouterr().out


def test_check_captures_w101(tmpdir, capsys):
    os.chdir(tmpdir)
    tmpdir.join("foo.py").write(
        """\
foo(x="foo"
"bar")
"""
    )
    res = check_file("foo.py", verbose=False, disabled_codes=set(), enabled_codes=set())
    assert res is False

    assert (
        "foo.py:1: unparenthesized multiline string concat in keyword arg (W101)"
        in capsys.readouterr().out
    )


def test_check_captures_w102(tmpdir, capsys):
    os.chdir(tmpdir)
    tmpdir.join("foo.py").write(
        """\
foo = {
    "x": "foo"
"bar"
}
"""
    )
    res = check_file("foo.py", verbose=False, disabled_codes=set(), enabled_codes=set())
    assert res is False

    assert (
        "foo.py:2: unparenthesized multiline string concat in dict value (W102)"
        in capsys.readouterr().out
    )


def test_check_captures_w103(tmpdir, capsys):
    os.chdir(tmpdir)
    tmpdir.join("foo.py").write(
        """\
foo = (
    "alpha",
    "beta"
    "gamma",
)
"""
    )
    res = check_file("foo.py", verbose=False, disabled_codes=set(), enabled_codes=set())
    assert res is False

    assert (
        "foo.py:3: unparenthesized multiline string concat in collection type (W103)"
        in capsys.readouterr().out
    )


def test_check_w103_ignores_solo_strings(tmpdir):
    os.chdir(tmpdir)
    tmpdir.join("foo.py").write(
        """\
foo = [
    "alpha"
    "beta"
]
"""
    )
    res = check_file("foo.py", verbose=False, disabled_codes=set(), enabled_codes=set())
    assert res is True


def test_check_captures_w200(tmpdir, capsys):
    os.chdir(tmpdir)
    tmpdir.join("foo.py").write(
        """\
def foo():
    if bar():
        return baz(quux("snork"))
    else:
        return baz(quux("snork"))
"""
    )
    res = check_file("foo.py", verbose=False, disabled_codes=set(), enabled_codes=set())
    assert res is False

    assert (
        "foo.py:2: two AST branches have identical contents (W200)"
        in capsys.readouterr().out
    )


def test_check_captures_w201(tmpdir, capsys):
    os.chdir(tmpdir)
    tmpdir.join("foo.py").write(
        """\
def foo():
    if bar():
        return baz()
    else:
        return baz()
"""
    )
    res = check_file("foo.py", verbose=False, disabled_codes=set(), enabled_codes=set())
    assert res is False

    assert (
        "foo.py:2: two AST branches have identical trivial contents (W201)"
        in capsys.readouterr().out
    )


def test_check_captures_w202(tmpdir, capsys):
    os.chdir(tmpdir)
    tmpdir.join("foo.py").write(
        """\
def foo():
    if bar():
        return baz("snork", flip="flop")
    elif qux():
        return quux()
    else:
        return baz("snork", flip="flop")
"""
    )
    res = check_file("foo.py", verbose=False, disabled_codes=set(), enabled_codes=set())
    assert res is False

    assert (
        "foo.py:2: two non-adjacent AST branches have identical contents (W202)"
        in capsys.readouterr().out
    )


def test_check_captures_w203(tmpdir, capsys):
    os.chdir(tmpdir)
    tmpdir.join("foo.py").write(
        """\
def foo():
    if bar():
        return baz("snork")
    elif qux():
        return quux()
    else:
        return baz("snork")
"""
    )
    res = check_file("foo.py", verbose=False, disabled_codes=set(), enabled_codes=set())
    assert res is False

    assert (
        "foo.py:2: two non-adjacent AST branches have identical trivial contents (W203)"
        in capsys.readouterr().out
    )


def test_can_disable_code_with_comment(tmpdir, capsys):
    os.chdir(tmpdir)
    tmpdir.join("foo.py").write(
        """\
def foo():
    x = "a " "b"  # slyp: disable=E100
"""
    )
    res = check_file("foo.py", verbose=False, disabled_codes=set(), enabled_codes=set())
    assert res is True

    assert "(E100)" not in capsys.readouterr().out


def test_can_disable_code_via_category_disable(tmpdir, capsys):
    os.chdir(tmpdir)
    tmpdir.join("foo.py").write(
        """\
def foo():
    if bar():
        return baz("snork")
    elif qux():
        return quux()
    else:
        return baz("snork")
"""
    )
    res = check_file(
        "foo.py", verbose=False, disabled_codes=set("W"), enabled_codes=set()
    )
    assert res is True

    assert "(W203)" not in capsys.readouterr().out


def test_check_captures_e110(tmpdir, capsys):
    os.chdir(tmpdir)
    tmpdir.join("foo.py").write(
        """\
if foo is None:
    return foo
"""
    )
    res = check_file("foo.py", verbose=False, disabled_codes=set(), enabled_codes=set())
    assert res is False

    assert (
        "foo.py:1: returning a variable checked as None, "
        "rather than returning None (E110)"
    ) in capsys.readouterr().out


def test_check_captures_w120(tmpdir, capsys):
    os.chdir(tmpdir)
    tmpdir.join("foo.py").write(
        """\
def foo(
    x: int |
    str
):
    return x
"""
    )
    res = check_file("foo.py", verbose=False, disabled_codes=set(), enabled_codes=set())
    assert res is False

    assert (
        "foo.py:2: unparenthesized multiline union in parameter annotation (W120)"
    ) in capsys.readouterr().out



from __future__ import annotations

import libcst

from ._base import ErrorCollectingVisitor
from .annotation_wrap import AnnotationWrapErrorCollector
from .str_concat import StrConcatErrorCollector

_VISITORS: list[ErrorCollectingVisitor] = [
    AnnotationWrapErrorCollector(),
    StrConcatErrorCollector(),
]


def run_cst_checkers(filename: str, content: bytes) -> set[tuple[int, str]]:
    try:
        tree = libcst.parse_module(content)
    except libcst.ParserSyntaxError:
        return {(0, "X001")}
    wrapper = libcst.MetadataWrapper(tree)
    for visitor in _VISITORS:
        visitor.filename = filename
        wrapper.visit(visitor)
    return {
        (lineno, code)
        for visitor in _VISITORS
        for (lineno, error_filename, code) in visitor.errors
        if error_filename == filename
    }


def _clear_visitor_errors() -> None:
    for visitor in _VISITORS:
        visitor.errors = set()



from __future__ import annotations

import libcst
import libcst.matchers

from ._base import ErrorCollectingVisitor

UNPARENTHESIZED_MULTILINE_CONCATENATED_STRING_MATCHER = libcst.matchers.ConcatenatedString(  # noqa: E501
    whitespace_between=(
        libcst.matchers.SimpleWhitespace(
            value=libcst.matchers.MatchIfTrue(lambda x: "\n" in x)
        )
        | libcst.matchers.ParenthesizedWhitespace()
    ),
    # empty list means that there are no parens
    # because `lpar` and `rpar` are `Sequence[Paren]` types
    lpar=[],
    rpar=[],
)

# SimpleWhitespace defines whitespace as seen between tokens in most contexts
# it can contain newlines *if* there is a preceding backslash escape
SIMPLE_WHITESPACE_NO_NEWLINE_MATCHER = libcst.matchers.SimpleWhitespace(
    value=libcst.matchers.MatchIfTrue(lambda x: "\n" not in x)
)

# check for 'unparenthesized multiline string concat in keyword arg'
# inside of function call sites
#
# the scenario here is that the string concat happens across multiple
# lines, but without parenthesization
# with a keyword argument getting the value, this is valid but results
# in a string with different physical indentation levels in the file
# e.g.
#   foo(
#       bar="alpha "
#       "beta"
#   )
#
# a classic case where this arises naturally is `help=...` for
# argparse and click, where the help string is often slightly too long
# for a single line, and easy to "break in two" without adding parens to
# force the whole block to indent
W101_KWARG_MATCHER = libcst.matchers.Arg(
    keyword=libcst.matchers.Name(),
    value=UNPARENTHESIZED_MULTILINE_CONCATENATED_STRING_MATCHER,
)

# check for 'unparenthesized multiline string concat in container'
#
# this case is multiline str concat inside of a container type, containing other
# elements
#
# the main scenario is "accidental multiline string", where a collection contains
# string literals, but is missing one comma
#
#   x = [
#       "foo "
#       "bar",
#       "baz",
#   ]
#
# however, a similar argument applies to any mixed iterable, e.g.
#
#   x = (
#       "foo "
#       "bar",
#       baz(),
#   ]
W103_ELEMENT_MATCHER = libcst.matchers.Element(
    value=UNPARENTHESIZED_MULTILINE_CONCATENATED_STRING_MATCHER
)

# this allows us to do faster matching for W103 by checking the whole list with
# a single matcher conditional
W103_IN_ELEMENT_LIST_MATCHER = libcst.matchers.OneOf(
    [
        libcst.matchers.AtLeastN(n=1),
        W103_ELEMENT_MATCHER,
        libcst.matchers.ZeroOrMore(),
    ],
    [
        libcst.matchers.ZeroOrMore(),
        W103_ELEMENT_MATCHER,
        libcst.matchers.AtLeastN(n=1),
    ],
)


class StrConcatErrorCollector(ErrorCollectingVisitor):
    METADATA_DEPENDENCIES = (libcst.metadata.PositionProvider,)

    def visit_ConcatenatedString(self, node: libcst.ConcatenatedString) -> None:
        # check for 'unnecessary string concat' situations
        # e.g.
        #   x = "foo " "bar"
        #
        # these are easily introduced when strings change in length and `black` is run
        # also common when `black` runs for the first time on a project
        if libcst.matchers.matches(
            node.whitespace_between, SIMPLE_WHITESPACE_NO_NEWLINE_MATCHER
        ):
            lpos = self.get_metadata(libcst.metadata.PositionProvider, node.left).start
            self.errors.add((lpos.line, self.filename, "E100"))

    def visit_Arg(self, node: libcst.Arg) -> None:
        if libcst.matchers.matches(node, W101_KWARG_MATCHER):
            lpos = self.get_metadata(
                libcst.metadata.PositionProvider,
                node.value.left,  # type: ignore[attr-defined]
            ).start
            self.errors.add((lpos.line, self.filename, "W101"))

    def visit_DictElement(self, node: libcst.DictElement) -> None:
        if libcst.matchers.matches(
            node.value,
            UNPARENTHESIZED_MULTILINE_CONCATENATED_STRING_MATCHER,
        ):
            lpos = self.get_metadata(
                libcst.metadata.PositionProvider,
                node.value.left,  # type: ignore[attr-defined]
            ).start
            self.errors.add((lpos.line, self.filename, "W102"))

    def visit_Tuple(self, node: libcst.Tuple) -> None:
        if libcst.matchers.matches(
            node,
            libcst.matchers.Tuple(
                elements=W103_IN_ELEMENT_LIST_MATCHER  # type: ignore[arg-type]
            ),
        ):
            self._collect_w103(node)

    def visit_List(self, node: libcst.List) -> None:
        if libcst.matchers.matches(
            node,
            libcst.matchers.List(
                elements=W103_IN_ELEMENT_LIST_MATCHER  # type: ignore[arg-type]
            ),
        ):
            self._collect_w103(node)

    def visit_Set(self, node: libcst.Set) -> None:
        if libcst.matchers.matches(
            node,
            libcst.matchers.Set(
                elements=W103_IN_ELEMENT_LIST_MATCHER  # type: ignore[arg-type]
            ),
        ):
            self._collect_w103(node)

    def _collect_w103(self, node: libcst.Tuple | libcst.List | libcst.Set) -> None:
        for element in node.elements:
            if libcst.matchers.matches(element, W103_ELEMENT_MATCHER):
                lpos = self.get_metadata(
                    libcst.metadata.PositionProvider,
                    element.value.left,  # type: ignore[attr-defined]
                ).start
                self.errors.add((lpos.line, self.filename, "W103"))
